import unittest

from pathlib import Path
import tempfile
import networkx as nx
from functools import partial
from collections import Counter

from alpaca import ProvenanceGraph, alpaca_setting


class ProvenanceGraphTestCase(unittest.TestCase):

    @staticmethod
    def _attr_comparison(attr_G1, attr_G2):
        return attr_G1 == attr_G2

    @staticmethod
    def _gexf_edge_comparison(attr_G1, attr_G2):
        # GEXF files add an `id` attribute to the edge, ignore it
        if 'id' in attr_G1:
            attr_G1.pop('id')
        if 'id' in attr_G2:
            attr_G2.pop('id')
        return attr_G1 == attr_G2

    @classmethod
    def setUpClass(cls):
        cls.ttl_path = Path(__file__).parent / "res"
        cls.temp_dir = tempfile.TemporaryDirectory(dir=cls.ttl_path,
                                                   suffix="tmp")
        cls.graph_comparison = partial(nx.is_isomorphic,
                                       node_match=cls._attr_comparison)
        alpaca_setting('authority', "my-authority")

    def test_graph_behavior_and_serialization(self):
        input_file = self.ttl_path / "input_output.ttl"
        graph = ProvenanceGraph(input_file)
        self.assertIsInstance(graph.graph, nx.DiGraph)
        self.assertEqual(len(graph.graph.nodes), 3)
        self.assertEqual(len(graph.graph.edges), 2)

        output_path = Path(self.temp_dir.name)

        # Test GEXF serialization
        gexf_file = output_path / "test.gexf"
        graph.save_gexf(gexf_file)
        self.assertTrue(gexf_file.exists())
        read_gexf = nx.read_gexf(gexf_file)
        self.assertTrue(self.graph_comparison(
            read_gexf, graph.graph, edge_match=self._gexf_edge_comparison))

        # Test GraphML serialization
        graphml_file = output_path / "test.graphml"
        graph.save_graphml(graphml_file)
        self.assertTrue(graphml_file.exists())
        read_graphml = nx.read_graphml(graphml_file)
        self.assertTrue(self.graph_comparison(read_graphml, graph.graph,
                                              edge_match=self._attr_comparison))

    def test_use_name_in_parameter(self):
        node = "urn:fz-juelich.de:alpaca:function_execution:Python:111111:999999:test.test_function#12345"
        input_file = self.ttl_path / "input_output.ttl"
        graph_use_name = ProvenanceGraph(input_file)
        graph_dont_use_name = ProvenanceGraph(input_file,
                                              use_name_in_parameter=False)
        use_name_attrs = graph_use_name.graph.nodes[node]
        dont_use_name_attrs = graph_dont_use_name.graph.nodes[node]

        self.assertTrue('test_function:param_1' in use_name_attrs)
        self.assertTrue('parameter:param_1' in dont_use_name_attrs)
        self.assertFalse('parameter:param_1' in use_name_attrs)
        self.assertFalse('test_function:param_1' in dont_use_name_attrs)
        self.assertEqual(use_name_attrs['test_function:param_1'], '5')
        self.assertEqual(dont_use_name_attrs['parameter:param_1'], '5')

    def test_remove_none(self):
        node = "urn:fz-juelich.de:alpaca:object:Python:builtins.NoneType:777777"
        input_file = self.ttl_path / "file_output.ttl"
        graph_with_none = ProvenanceGraph(input_file, remove_none=False)
        graph_without_none = ProvenanceGraph(input_file, remove_none=True)

        self.assertTrue(node in graph_with_none.graph.nodes)
        self.assertFalse(node in graph_without_none.graph.nodes)
        self.assertEqual(len(graph_with_none.graph.nodes), 4)
        self.assertEqual(len(graph_without_none.graph.nodes), 3)

    def test_remove_none_no_output_function(self):
        node = "urn:fz-juelich.de:alpaca:object:Python:builtins.NoneType:777777"
        input_file = self.ttl_path / "none_output.ttl"
        graph_with_none = ProvenanceGraph(input_file, remove_none=False)
        graph_without_none = ProvenanceGraph(input_file, remove_none=True)

        self.assertTrue(node in graph_with_none.graph.nodes)
        self.assertFalse(node in graph_without_none.graph.nodes)
        self.assertEqual(len(graph_with_none.graph.nodes), 3)
        self.assertEqual(len(graph_without_none.graph.nodes), 2)


    def test_memberships(self):
        input_file = self.ttl_path / "multiple_memberships.ttl"
        graph = ProvenanceGraph(input_file)

        self.assertEqual(len(graph.graph.nodes), 7)
        for node in (
        "urn:fz-juelich.de:alpaca:object:Python:builtins.list:3333332",
        "urn:fz-juelich.de:alpaca:object:Python:builtins.list:23333332",
        "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333",
        "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345",
        "urn:fz-juelich.de:alpaca:object:Python:test.SuperContainer:2333333"):
            self.assertTrue(node in graph.graph.nodes)

        for expected_edge, expected_label in zip(
                ((
                 "urn:fz-juelich.de:alpaca:object:Python:test.SuperContainer:2333333",
                 "urn:fz-juelich.de:alpaca:object:Python:builtins.list:23333332"),
                 (
                 "urn:fz-juelich.de:alpaca:object:Python:builtins.list:23333332",
                 "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333"),
                 (
                 "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333",
                 "urn:fz-juelich.de:alpaca:object:Python:builtins.list:3333332"),
                 (
                 "urn:fz-juelich.de:alpaca:object:Python:builtins.list:3333332",
                 "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345")),
                (".containers", "[0]", ".inputs", "[1]")):
            self.assertTrue(expected_edge in graph.graph.edges)
            self.assertTrue(graph.graph.edges[expected_edge]['membership'])
            self.assertEqual(graph.graph.edges[expected_edge]['label'],
                             expected_label)

    def test_membership_condensing(self):
        input_file = self.ttl_path / "multiple_memberships.ttl"
        graph = ProvenanceGraph(input_file)
        graph.condense_memberships()

        self.assertEqual(len(graph.graph.nodes), 4)

        for node in (
        "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345",
        "urn:fz-juelich.de:alpaca:object:Python:test.SuperContainer:2333333"):
            self.assertTrue(node in graph.graph.nodes)

        for node in (
        "urn:fz-juelich.de:alpaca:object:Python:builtins.list:3333332",
        "urn:fz-juelich.de:alpaca:object:Python:builtins.list:23333332",
        "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333"):
            self.assertFalse(node in graph.graph.nodes)

        for edge in ((
                     "urn:fz-juelich.de:alpaca:object:Python:test.SuperContainer:2333333",
                     "urn:fz-juelich.de:alpaca:object:Python:builtins.list:23333332"),
                     (
                     "urn:fz-juelich.de:alpaca:object:Python:builtins.list:23333332",
                     "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333"),
                     (
                     "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333",
                     "urn:fz-juelich.de:alpaca:object:Python:builtins.list:3333332"),
                     (
                     "urn:fz-juelich.de:alpaca:object:Python:builtins.list:3333332",
                     "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345")):
            self.assertFalse(edge in graph.graph.edges)

        expected_edge = (
        "urn:fz-juelich.de:alpaca:object:Python:test.SuperContainer:2333333",
        "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345")

        expected_label = ".containers[0].inputs[1]"

        self.assertTrue(expected_edge in graph.graph.edges)
        self.assertTrue(graph.graph.edges[expected_edge]['membership'])
        self.assertEqual(graph.graph.edges[expected_edge]['label'],
                         expected_label)

    def test_membership_condensing_with_preservation(self):
        input_file = self.ttl_path / "multiple_memberships.ttl"
        graph = ProvenanceGraph(input_file)
        graph.condense_memberships(preserve=['Container'])

        self.assertEqual(len(graph.graph.nodes), 5)

        for node in (
        "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345",
        "urn:fz-juelich.de:alpaca:object:Python:test.SuperContainer:2333333",
        "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333"):
            self.assertTrue(node in graph.graph.nodes)

        for node in (
        "urn:fz-juelich.de:alpaca:object:Python:builtins.list:3333332",
        "urn:fz-juelich.de:alpaca:object:Python:builtins.list:23333332"):
            self.assertFalse(node in graph.graph.nodes)

        for edge in ((
                     "urn:fz-juelich.de:alpaca:object:Python:test.SuperContainer:2333333",
                     "urn:fz-juelich.de:alpaca:object:Python:builtins.list:23333332"),
                     (
                     "urn:fz-juelich.de:alpaca:object:Python:builtins.list:23333332",
                     "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333"),
                     (
                     "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333",
                     "urn:fz-juelich.de:alpaca:object:Python:builtins.list:3333332"),
                     (
                     "urn:fz-juelich.de:alpaca:object:Python:builtins.list:3333332",
                     "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345")):
            self.assertFalse(edge in graph.graph.edges)

        for expected_edge, expected_label in zip(
                ((
                 "urn:fz-juelich.de:alpaca:object:Python:test.SuperContainer:2333333",
                 "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333"),
                 (
                 "urn:fz-juelich.de:alpaca:object:Python:test.Container:333333",
                 "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345")),
                (".containers[0]", ".inputs[1]")):
            self.assertTrue(expected_edge in graph.graph.edges)
            self.assertTrue(graph.graph.edges[expected_edge]['membership'])
            self.assertEqual(graph.graph.edges[expected_edge]['label'],
                             expected_label)

    def test_strip_namespace(self):
        input_file = self.ttl_path / "metadata.ttl"

        graph = ProvenanceGraph(input_file, attributes=['name'],
                                annotations=['sua'], strip_namespace=True)
        node_attrs = graph.graph.nodes[
            "urn:fz-juelich.de:alpaca:object:Python:neo.core.SpikeTrain:54321"]

        self.assertEqual(node_attrs["name"], "Spiketrain#1")
        self.assertEqual(node_attrs["sua"], "false")

    def test_use_class_in_method_name(self):
        input_file = self.ttl_path / "class_method.ttl"

        graph = ProvenanceGraph(input_file, attributes=None,
                                annotations=None,
                                use_class_in_method_name=True)
        node_attrs = graph.graph.nodes[
            "urn:fz-juelich.de:alpaca:function_execution:Python:111111:999999:test.ObjectWithMethod.process#12345"]

        self.assertEqual(node_attrs["label"], "ObjectWithMethod.process")

    def test_no_use_class_in_method_name(self):
        input_file = self.ttl_path / "class_method.ttl"

        graph = ProvenanceGraph(input_file, attributes=None,
                                annotations=None,
                                use_class_in_method_name=False)
        node_attrs = graph.graph.nodes[
            "urn:fz-juelich.de:alpaca:function_execution:Python:111111:999999:test.ObjectWithMethod.process#12345"]

        self.assertEqual(node_attrs["label"], "process")

    def test_no_strip_namespace(self):
        input_file = self.ttl_path / "metadata.ttl"

        graph = ProvenanceGraph(input_file, attributes=['name'],
                                annotations=['sua'], strip_namespace=False)
        node_attrs = graph.graph.nodes[
            "urn:fz-juelich.de:alpaca:object:Python:neo.core.SpikeTrain:54321"]

        self.assertEqual(node_attrs["attribute:name"], "Spiketrain#1")
        self.assertEqual(node_attrs["annotation:sua"], "false")

    def test_attributes(self):
        input_file = self.ttl_path / "metadata.ttl"
        graph = ProvenanceGraph(input_file, attributes=['metadata_2'])

        node_attrs = graph.graph.nodes[
            "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345"]

        self.assertEqual(node_attrs["metadata_2"], "5")

        for annotation in ("metadata_1", "metadata_3", "metadata_4"):
            self.assertTrue(annotation not in node_attrs)

        node_attrs = graph.graph.nodes[
            "urn:fz-juelich.de:alpaca:object:Python:neo.core.SpikeTrain:54321"]

        self.assertTrue("name" not in node_attrs)

    def test_annotations(self):
        input_file = self.ttl_path / "metadata.ttl"
        graph = ProvenanceGraph(input_file, annotations=['sua'])

        node_attrs = graph.graph.nodes[
            "urn:fz-juelich.de:alpaca:object:Python:neo.core.SpikeTrain:54321"]

        self.assertEqual(node_attrs['sua'], "false")

        for annotation in ("channel", "complexity", "event"):
            self.assertTrue(annotation not in node_attrs)

    def test_all_annotations(self):
        input_file = self.ttl_path / "metadata.ttl"
        graph = ProvenanceGraph(input_file, annotations='all')

        annotations_node = "urn:fz-juelich.de:alpaca:object:Python:neo.core.SpikeTrain:54321"

        expected_annotations = {"sua": "false",
                                "channel": "56",
                                "complexity": "[0 1 2 3]",
                                "event": "[ True False False]"}

        attributes_node = "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345"

        expected_attributes = {"metadata_1": "value1",
                               "metadata_2": "5",
                               "metadata_3": "5.0",
                               "metadata_4": "true"}

        node_attrs = graph.graph.nodes[annotations_node]
        for key, value in expected_annotations.items():
            self.assertTrue(key in node_attrs)
            self.assertEqual(node_attrs[key], value)

        node_attrs = graph.graph.nodes[attributes_node]
        for key, value in expected_attributes.items():
            self.assertTrue(key not in node_attrs)

    def test_all_attributes(self):
        input_file = self.ttl_path / "metadata.ttl"
        graph = ProvenanceGraph(input_file, attributes='all')

        annotations_node = "urn:fz-juelich.de:alpaca:object:Python:neo.core.SpikeTrain:54321"

        expected_annotations = {"sua": "false",
                                "channel": "56",
                                "complexity": "[0 1 2 3]",
                                "event": "[ True False False]"}

        attributes_node = "urn:fz-juelich.de:alpaca:object:Python:test.InputObject:12345"

        expected_attributes = {"metadata_1": "value1",
                               "metadata_2": "5",
                               "metadata_3": "5.0",
                               "metadata_4": "true"}

        node_attrs = graph.graph.nodes[annotations_node]
        for key, value in expected_annotations.items():
            self.assertTrue(key not in node_attrs)

        self.assertEqual(node_attrs['name'], "Spiketrain#1")

        node_attrs = graph.graph.nodes[attributes_node]
        for key, value in expected_attributes.items():
            self.assertTrue(key in node_attrs)
            self.assertEqual(node_attrs[key], value)


class GraphAggregationTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ttl_path = Path(__file__).parent / "res"
        input_file = cls.ttl_path / "parallel_graph.ttl"
        cls.graph = ProvenanceGraph(input_file, attributes=['shape', 'metadata'])
        alpaca_setting('authority', "my-authority")

    def test_serialization(self):
        temp_dir = tempfile.TemporaryDirectory(dir=self.ttl_path, suffix="tmp")
        output_path = Path(temp_dir.name)

        gexf_file = output_path / "test.gexf"
        self.graph.aggregate({}, output_file=gexf_file)
        self.assertTrue(gexf_file.exists())

        graphml_file = output_path / "test.graphml"
        self.graph.aggregate({}, output_file=graphml_file)
        self.assertTrue(graphml_file.exists())

        with self.assertRaises(ValueError):
            self.graph.aggregate({}, output_file=output_path / "test.invalid")

    def test_overall_aggregation(self):
        aggregated = self.graph.aggregate({}, use_function_parameters=False,
                                          output_file=None)
        nodes = aggregated.nodes

        self.assertEqual(len(nodes), 4)

        expected_values_per_node = {
            'OutputObject': {'metadata': "0;1",
                             'shape': "(2,);(3,);(4,);(5,)"},
            'InputObject': {'metadata': "5",
                            'shape': "(2,);(3,);(4,);(5,)"},
            'process': {'process:value': "0;1;2;3"},
            'list': {}
        }

        all_labels = [nodes[node]['label'] for node in nodes]
        counts = Counter(all_labels)
        self.assertEqual(counts['OutputObject'], 1)
        self.assertEqual(counts['InputObject'], 1)
        self.assertEqual(counts['process'], 1)
        self.assertEqual(counts['list'], 1)

        for node, attrs in nodes.items():
            label = attrs['label']
            with self.subTest(f"Node label {label}"):
                self.assertTrue(label in expected_values_per_node)
                for key, value in expected_values_per_node[label].items():
                    self.assertEqual(attrs[key], value)

    def test_aggregation_by_attribute(self):
        aggregated = self.graph.aggregate({'InputObject': ('shape',)},
                                          use_function_parameters=False,
                                          output_file=None)
        nodes = aggregated.nodes

        self.assertEqual(len(nodes), 7)

        expected_values_per_node = {
            'OutputObject': {'metadata': "0;1",
                             'shape': "(2,);(3,);(4,);(5,)"},
            'InputObject': {'metadata': "5",
                            'shape': ["(2,)", "(3,)", "(4,)", "(5,)"]},
            'process': {'process:value': "0;1;2;3"},
            'list': {}
        }

        all_labels = [nodes[node]['label'] for node in nodes]
        counts = Counter(all_labels)
        self.assertEqual(counts['OutputObject'], 1)
        self.assertEqual(counts['InputObject'], 4)
        self.assertEqual(counts['process'], 1)
        self.assertEqual(counts['list'], 1)

        for node, attrs in nodes.items():
            label = attrs['label']
            with self.subTest(f"Node label {label}"):
                self.assertTrue(label in expected_values_per_node)
                for key, value in expected_values_per_node[label].items():
                    if not isinstance(value, list):
                        self.assertEqual(attrs[key], value)
                    else:
                        self.assertTrue(attrs[key] in value)

    def test_aggregation_by_attribute_with_function(self):
        aggregated = self.graph.aggregate({'InputObject': ('shape',)},
                                          use_function_parameters=True,
                                          output_file=None)
        nodes = aggregated.nodes

        self.assertEqual(len(nodes), 10)

        expected_values_per_node = {
            'OutputObject': {'metadata': "0;1",
                             'shape': "(2,);(3,);(4,);(5,)"},
            'InputObject': {'metadata': "5",
                            'shape': ["(2,)", "(3,)", "(4,)", "(5,)"]},
            'process': {'process:value': ["0", "1", "2", "3"]},
            'list': {}
        }

        all_labels = [nodes[node]['label'] for node in nodes]
        counts = Counter(all_labels)
        self.assertEqual(counts['OutputObject'], 1)
        self.assertEqual(counts['InputObject'], 4)
        self.assertEqual(counts['process'], 4)
        self.assertEqual(counts['list'], 1)

        for node, attrs in nodes.items():
            label = attrs['label']
            with self.subTest(f"Node label {label}"):
                self.assertTrue(label in expected_values_per_node)
                for key, value in expected_values_per_node[label].items():
                    if not isinstance(value, list):
                        self.assertEqual(attrs[key], value)
                    else:
                        self.assertTrue(attrs[key] in value)


if __name__ == "__main__":
    unittest.main()



import unittest

import joblib
import datetime
import sys
from io import StringIO
import tempfile
from pathlib import Path

import rdflib
from rdflib.compare import graph_diff

import numpy as np

from alpaca import (Provenance, activate, deactivate, save_provenance,
                    print_history)
from alpaca.alpaca_types import (FunctionInfo, Container, DataObject, File)

# Define some data and expected values test tracking

TEST_ARRAY = np.array([1, 2, 3])
TEST_ARRAY_INFO = DataObject(hash=joblib.hash(TEST_ARRAY, hash_name='sha1'),
                             hash_method="joblib_SHA1",
                             type="numpy.ndarray", id=id(TEST_ARRAY),
                             details={'shape': (3,), 'dtype': np.int64})

TEST_ARRAY_2 = np.array([4, 5, 6])
TEST_ARRAY_2_INFO = DataObject(hash=joblib.hash(TEST_ARRAY_2,
                                                hash_name='sha1'),
                               hash_method="joblib_SHA1",
                               type="numpy.ndarray", id=id(TEST_ARRAY_2),
                               details={'shape': (3,), 'dtype': np.int64})

CONTAINER = [TEST_ARRAY, TEST_ARRAY_2]


# Define some functions to test tracking in different scenarios

@Provenance(inputs=['array'])
def simple_function(array, param1, param2):
    """ Takes a single input and outputs a single element"""
    return array + 3


@Provenance(inputs=['array'])
def simple_function_default(array, param1, param2=10):
    """ Takes a single input and outputs a single element.
    One kwarg is default
    """
    return array + 5


@Provenance(inputs=None, container_input=['arrays'])
def container_input_function(arrays, param1, param2):
    """ Takes a container input (e.g. list) and outputs a single element"""
    return np.mean(arrays)


@Provenance(inputs=['arrays'])
def varargs_function(*arrays, param1, param2):
    """ Takes a variable argument input and outputs a single element"""
    return np.mean(arrays)


@Provenance(inputs=['array'])
def multiple_outputs_function(array, param1, param2):
    """ Takes a single input and outputs multiple elements as a tuple"""
    return array + 3, array + 4


@Provenance(inputs=['array_1', 'array_2'])
def multiple_inputs_function(array_1, array_2, param1, param2):
    """ Takes multiple inputs and outputs a single element"""
    return array_1 + array_2


@Provenance(inputs=['array'], container_output=True)
def container_output_function(array, param1, param2):
    """ Takes a single input and outputs multiple elements in a container"""
    return [array + i for i in range(3, 5)]


# Function to help verifying FunctionExecution tuples
def _check_function_execution(actual, exp_function, exp_input, exp_params,
                              exp_output, exp_arg_map, exp_kwarg_map,
                              exp_code_stmnt, exp_return_targets, exp_order,
                              test_case):
    # Check function
    test_case.assertTupleEqual(actual.function, exp_function)

    # Check inputs
    for input_arg, value in actual.input.items():
        test_case.assertTrue(input_arg in exp_input)
        test_case.assertTupleEqual(value, exp_input[input_arg])

    # Check parameters
    test_case.assertDictEqual(actual.params, exp_params)

    # Check outputs
    for output, value in actual.output.items():
        test_case.assertTrue(output in exp_output)
        test_case.assertTupleEqual(value, exp_output[output])

    # Check args and kwargs
    test_case.assertListEqual(actual.arg_map, exp_arg_map)
    test_case.assertListEqual(actual.kwarg_map, exp_kwarg_map)

    # Check other information
    test_case.assertEqual(actual.code_statement, exp_code_stmnt)
    test_case.assertListEqual(actual.return_targets, exp_return_targets)
    test_case.assertEqual(actual.order, exp_order)
    test_case.assertNotEqual(actual.execution_id, "")

    # Check time stamps are valid ISO dates
    test_case.assertIsInstance(
        datetime.datetime.fromisoformat(actual.time_stamp_start),
        datetime.datetime)
    test_case.assertIsInstance(
        datetime.datetime.fromisoformat(actual.time_stamp_end),
        datetime.datetime)


class ProvenanceDecoratorInterfaceFunctionsTestCase(unittest.TestCase):

    def test_activate_deactivate(self):
        activate(clear=True)
        simple_function(TEST_ARRAY, 1, 2)
        simple_function(TEST_ARRAY, 3, 4)
        deactivate()
        simple_function(TEST_ARRAY, 5, 6)

        self.assertEqual(len(Provenance.history), 2)
        self.assertEqual(Provenance.history[0].code_statement,
                         "simple_function(TEST_ARRAY, 1, 2)")
        self.assertEqual(Provenance.history[1].code_statement,
                         "simple_function(TEST_ARRAY, 3, 4)")

    def test_save_provenance(self):
        activate(clear=True)
        res = simple_function(TEST_ARRAY, 1, 2)
        deactivate()

        # For every supported format, serialize to a string
        for output_format in ('json-ld', 'n3', 'nt', 'hext', 'pretty-xml',
                              'trig', 'turtle', 'longturtle', 'xml'):
            with self.subTest(f"Serialization format",
                              output_format=output_format):
                serialization = save_provenance(file_name=None,
                                                file_format=output_format)
                self.assertNotEqual(serialization, "")

        # For shortucts, test the expected serializations
        for short_output_format, output_format in (('ttl', 'turtle'),
                                                   ('rdf', 'xml'),
                                                   ('json', 'json-ld')):
            with self.subTest(f"Short serialization format",
                              short_format=short_output_format):
                short = save_provenance(None, file_format=short_output_format)
                serialization = save_provenance(None, file_format=output_format)

                short_graph = rdflib.Graph()
                short_graph.parse(StringIO(short), format=output_format)

                serialization_graph = rdflib.Graph()
                serialization_graph.parse(StringIO(serialization),
                                          format=output_format)

                self.assertTrue(short_graph.isomorphic(serialization_graph))
                _, in_first, in_second = graph_diff(short_graph,
                                                    serialization_graph)
                self.assertEqual(len(in_first), 0)
                self.assertEqual(len(in_second), 0)

    def test_print_history(self):
        activate(clear=True)
        res = simple_function(TEST_ARRAY, 1, 2)
        deactivate()

        expected_str = str(Provenance.history)

        # Capture STDOUT and print
        captured = StringIO()
        sys.stdout = captured
        print_history()
        sys.stdout = sys.__stdout__

        self.assertEqual(captured.getvalue().replace("\n", ""), expected_str)


class ProvenanceDecoratorFunctionsTestCase(unittest.TestCase):

    def test_get_module_version(self):
        expected_numpy_version = np.__version__

        numpy_version = Provenance._get_module_version("numpy", "mean")
        self.assertEqual(numpy_version, expected_numpy_version)

        numpy_version_submodule = Provenance._get_module_version(
            "numpy.random", "normal")
        self.assertEqual(numpy_version_submodule, expected_numpy_version)

        main_version = Provenance._get_module_version("__main__",
                                                      "test_function")
        self.assertEqual(main_version, "")

        invalid = Provenance._get_module_version("non_existent", "test")
        self.assertEqual(invalid, "")


class ProvenanceDecoratorInputOutputCombinationsTestCase(unittest.TestCase):

    def test_simple_function(self):
        activate(clear=True)
        res = simple_function(TEST_ARRAY, 1, 2)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_output = DataObject(
            hash=joblib.hash(TEST_ARRAY+3, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res),
            details={'shape': (3,), 'dtype': np.int64})

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('simple_function',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 1, 'param2': 2},
            exp_output={0: expected_output},
            exp_arg_map=['array', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = simple_function(TEST_ARRAY, 1, 2)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_simple_function_no_target(self):
        activate(clear=True)
        simple_function(TEST_ARRAY, param2=1, param1=2)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        # In this test we cannot know the id of the output, as it is not
        # stored in any variable. Let's get it from the history so that the
        # test does not fail
        output_id = Provenance.history[0].output[0].id

        expected_output = DataObject(
            hash=joblib.hash(TEST_ARRAY+3, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=output_id,
            details={'shape': (3,), 'dtype': np.int64})

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('simple_function',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 2, 'param2': 1},
            exp_output={0: expected_output},
            exp_arg_map=['array'],
            exp_kwarg_map=['param1', 'param2'],
            exp_code_stmnt="simple_function(TEST_ARRAY, param2=1, param1=2)",
            exp_return_targets=[],
            exp_order=1,
            test_case=self)

    def test_kwargs_params(self):
        activate(clear=True)
        res = simple_function(TEST_ARRAY, 1, param2=2)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_output = DataObject(
            hash=joblib.hash(TEST_ARRAY+3, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res),
            details={'shape': (3,), 'dtype': np.int64})

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('simple_function',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 1, 'param2': 2},
            exp_output={0: expected_output},
            exp_arg_map=['array', 'param1'],
            exp_kwarg_map=['param2'],
            exp_code_stmnt="res = simple_function(TEST_ARRAY, 1, param2=2)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_kwargs_params_default(self):
        activate(clear=True)
        res = simple_function_default(TEST_ARRAY, 1)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_output = DataObject(
            hash=joblib.hash(TEST_ARRAY+5, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res),
            details={'shape': (3,), 'dtype': np.int64})

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('simple_function_default',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 1, 'param2': 10},
            exp_output={0: expected_output},
            exp_arg_map=['array', 'param1'],
            exp_kwarg_map=['param2'],
            exp_code_stmnt="res = simple_function_default(TEST_ARRAY, 1)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_kwargs_params_default_override(self):
        activate(clear=True)
        res = simple_function_default(TEST_ARRAY, 1, 8)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_output = DataObject(
            hash=joblib.hash(TEST_ARRAY+5, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res),
            details={'shape': (3,), 'dtype': np.int64})

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('simple_function_default',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 1, 'param2': 8},
            exp_output={0: expected_output},
            exp_arg_map=['array', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = simple_function_default(TEST_ARRAY, 1, 8)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_container_input_function(self):
        activate(clear=True)
        avg = container_input_function(CONTAINER, 3, 6)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)
        self.assertEqual(avg, 3.5)

        expected_output = DataObject(
            hash=joblib.hash(np.float64(3.5), hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.float64", id=id(avg),
            details={'shape': (), 'dtype': np.float64})

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('container_input_function',
                                      'test_decorator', ''),
            exp_input={'arrays': Container(tuple(
                [TEST_ARRAY_INFO, TEST_ARRAY_2_INFO]))},
            exp_params={'param1': 3, 'param2': 6},
            exp_output={0: expected_output},
            exp_arg_map=['arrays', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="avg = container_input_function(CONTAINER, 3, 6)",
            exp_return_targets=['avg'],
            exp_order=1,
            test_case=self)

    def test_varargs_input_function(self):
        activate(clear=True)
        avg = varargs_function(*CONTAINER, param1=1, param2=2)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)
        self.assertEqual(avg, 3.5)

        expected_output = DataObject(
            hash=joblib.hash(np.float64(3.5), hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.float64", id=id(avg),
            details={'shape': (), 'dtype': np.float64})

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('varargs_function',
                                      'test_decorator', ''),
            exp_input={'arrays': Container(tuple(
                [TEST_ARRAY_INFO, TEST_ARRAY_2_INFO]))},
            exp_params={'param1': 1, 'param2': 2},
            exp_output={0: expected_output},
            exp_arg_map=['arrays'],
            exp_kwarg_map=['param1', 'param2'],
            exp_code_stmnt="avg = varargs_function(*CONTAINER, param1=1, param2=2)",
            exp_return_targets=['avg'],
            exp_order=1,
            test_case=self)

    def test_multiple_inputs_function(self):
        activate(clear=True)
        res = multiple_inputs_function(TEST_ARRAY, TEST_ARRAY_2, 3, 6)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_output = DataObject(
            hash=joblib.hash(TEST_ARRAY+TEST_ARRAY_2, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res),
            details={'shape': (3,), 'dtype': np.int64})

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('multiple_inputs_function',
                                      'test_decorator', ''),
            exp_input={'array_1': TEST_ARRAY_INFO,
                       'array_2': TEST_ARRAY_2_INFO},
            exp_params={'param1': 3, 'param2': 6},
            exp_output={0: expected_output},
            exp_arg_map=['array_1', 'array_2', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = multiple_inputs_function(TEST_ARRAY, TEST_ARRAY_2, 3, 6)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_multiple_outputs_function_elements(self):
        activate(clear=True)
        res1, res2 = multiple_outputs_function(TEST_ARRAY, 3, 6)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_output_1 = DataObject(
            hash=joblib.hash(TEST_ARRAY+3, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res1),
            details={'shape': (3,), 'dtype': np.int64})

        expected_output_2 = DataObject(
            hash=joblib.hash(TEST_ARRAY+4, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res2),
            details={'shape': (3,), 'dtype': np.int64})

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('multiple_outputs_function',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 3, 'param2': 6},
            exp_output={0: expected_output_1, 1: expected_output_2},
            exp_arg_map=['array', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res1, res2 = multiple_outputs_function(TEST_ARRAY, 3, 6)",
            exp_return_targets=['res1', 'res2'],
            exp_order=1,
            test_case=self)

    def test_multiple_outputs_function_tuple(self):
        activate(clear=True)
        res = multiple_outputs_function(TEST_ARRAY, 3, 6)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_output = DataObject(
            hash=joblib.hash((TEST_ARRAY+3, TEST_ARRAY+4), hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="builtins.tuple", id=id(res),
            details={})

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('multiple_outputs_function',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 3, 'param2': 6},
            exp_output={0: expected_output},
            exp_arg_map=['array', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = multiple_outputs_function(TEST_ARRAY, 3, 6)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_container_output_function(self):
        activate(clear=True)
        res = container_output_function(TEST_ARRAY, 3, 6)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_output_1 = DataObject(
            hash=joblib.hash(TEST_ARRAY + 3, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res[0]),
            details={'shape': (3,), 'dtype': np.int64})

        expected_output_2 = DataObject(
            hash=joblib.hash(TEST_ARRAY + 4, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res[1]),
            details={'shape': (3,), 'dtype': np.int64})

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('container_output_function',
                                      'test_decorator', ''),
            exp_input={'array': TEST_ARRAY_INFO},
            exp_params={'param1': 3, 'param2': 6},
            exp_output={0: expected_output_1, 1: expected_output_2},
            exp_arg_map=['array', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = container_output_function(TEST_ARRAY, 3, 6)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

@Provenance(inputs=None, file_input=['file_name'])
def extract_words_from_file(file_name):
    with open(file_name, "r") as input_file:
        words = input_file.read().split(" ")
    return words


@Provenance(inputs=['words'], file_output=['file_name'])
def save_words_to_file(words, file_name):
    with open(file_name, "w") as output_file:
        output_file.writelines(" ".join(words))


class ProvenanceDecoratorFileInputOutputTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.res_path = Path(__file__).parent.absolute() / "res"

    def test_file_input(self):
        activate(clear=True)
        file_name = self.res_path / "file_input.txt"
        res = extract_words_from_file(file_name)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_list = ["This", "is", "an", "example", "file", "used", "as",
                         "input", "for", "the", "tests."]
        expected_output = DataObject(
            hash=joblib.hash(expected_list, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="builtins.list", id=id(res), details={})

        expected_file = File("96ccc1380e069667069acecea3e2ab559441657807e0a86d14f49028710ddb3a",
                             hash_type="sha256", path=file_name)

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('extract_words_from_file',
                                      'test_decorator', ''),
            exp_input={'file_name': expected_file},
            exp_params={},
            exp_output={0: expected_output},
            exp_arg_map=['file_name'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = extract_words_from_file(file_name)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)

    def test_file_output(self):
        activate(clear=True)
        tmp_file = tempfile.NamedTemporaryFile(dir=self.res_path, delete=True)
        file_name = tmp_file.name
        input_list = ["Some", "words", "were", "written", "to", "this", "file"]
        res = save_words_to_file(input_list, file_name)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        expected_input = DataObject(
            hash=joblib.hash(input_list, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="builtins.list", id=id(input_list), details={})

        # As None has its own UUID, let's get what was generated
        self.assertEqual(len(Provenance.history), 1)
        output_uuid = Provenance.history[0].output[0].hash

        expected_none_output = DataObject(hash=output_uuid, hash_method="UUID",
            type="builtins.NoneType", id=id(res), details={})

        expected_file = File("00d20b4831b0dadded2c633bdfc3dde3926fc17baaed51dacdab3e52a3b0d419",
                             hash_type="sha256", path=Path(file_name))

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('save_words_to_file',
                                      'test_decorator', ''),
            exp_input={'words': expected_input},
            exp_params={},
            exp_output={0: expected_none_output, 'file.0': expected_file},
            exp_arg_map=['words', 'file_name'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = save_words_to_file(input_list, file_name)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)


# Tracking methods inside classes
class ObjectWithMethod(object):
    def __init__(self, coefficient):
        self.coefficient = coefficient

    def process(self, array, param1, param2):
        return array + self.coefficient


ObjectWithMethod.process = Provenance(inputs=['self', 'array'])(
    ObjectWithMethod.process)


class ProvenanceDecoratorClassMethodsTestCase(unittest.TestCase):

    def test_class_method(self):
        activate(clear=True)
        obj = ObjectWithMethod(2)
        res = obj.process(TEST_ARRAY, 4, 5)
        deactivate()

        self.assertEqual(len(Provenance.history), 1)

        obj_info = DataObject(
            hash=joblib.hash(obj, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="test_decorator.ObjectWithMethod",
            id=id(obj),
            details={'coefficient': 2})

        expected_output = DataObject(
            hash=joblib.hash(TEST_ARRAY+2, hash_name='sha1'),
            hash_method="joblib_SHA1",
            type="numpy.ndarray", id=id(res),
            details={'shape': (3,), 'dtype': np.int64})

        _check_function_execution(
            actual=Provenance.history[0],
            exp_function=FunctionInfo('ObjectWithMethod.process',
                                      'test_decorator', ''),
            exp_input={'self': obj_info, 'array': TEST_ARRAY_INFO},
            exp_params={'param1': 4, 'param2': 5},
            exp_output={0: expected_output},
            exp_arg_map=['self', 'array', 'param1', 'param2'],
            exp_kwarg_map=[],
            exp_code_stmnt="res = obj.process(TEST_ARRAY, 4, 5)",
            exp_return_targets=['res'],
            exp_order=1,
            test_case=self)


if __name__ == "__main__":
    unittest.main()



"""
This class provides functionality to serialize/deserialize the provenance
using an ontology based on the W3C Provenance Ontology (PROV-O). The Alpaca
ontology is used to serialize the provenance information captured by Alpaca as
RDF files.

.. autoclass:: alpaca.AlpacaProvDocument
    :members:

"""

from itertools import product

from rdflib import Graph, URIRef, BNode, Literal
from rdflib.namespace import RDF, PROV, XSD

from alpaca.ontology import ALPACA
from alpaca.serialization.identifiers import (data_object_identifier,
                                              file_identifier,
                                              function_identifier,
                                              script_identifier,
                                              execution_identifier)
from alpaca.serialization.converters import _ensure_type
from alpaca.serialization.neo import _neo_object_metadata

from alpaca.utils.files import _get_prov_file_format
from alpaca.alpaca_types import DataObject, File, Container
from alpaca.settings import _ALPACA_SETTINGS


def _add_name_value_pair(graph, uri, predicate, name, value):
    # Add a relationship defined by `predicate` using a blank node as object.
    # The object will be of type `alpaca:NameValuePair`.
    blank_node = BNode()
    graph.add((uri, predicate, blank_node))
    graph.add((blank_node, RDF.type, ALPACA.NameValuePair))
    graph.add((blank_node, ALPACA.pairName, Literal(name)))
    graph.add((blank_node, ALPACA.pairValue, Literal(value)))


class AlpacaProvDocument(object):
    """
    Generates a file using the Alpaca ontology (based on W3C PROV-O) from
    the history records captured during the execution of a Python script,
    or reads a serialized file into an RDF graph object.

    Attributes
    ----------
    graph : rdflib.Graph
        Provenance data represented as an RDF graph, using the Alpaca ontology
        based on PROV-O.

    Notes
    -----
    For convenience, you can serialize the active history easily by just using
    the :func:`save_provenance` in :ref:`interface`. This class should
    be used only if you want to access the data as an RDF graph or to manually
    control the serialization.
    """

    def __init__(self):
        self.graph = Graph()
        self.graph.namespace_manager.bind('alpaca', ALPACA)
        self.graph.namespace_manager.bind('prov', PROV)
        self._authority = _ALPACA_SETTINGS['authority']

        # Metadata plugins are used for packages (e.g., Neo) that require
        # special handling of metadata when adding to the PROV records.
        # Plugins are external functions that take the graph, the object URI,
        # and the metadata dict as parameters.
        self._metadata_plugins = {
            'neo': _neo_object_metadata
        }

    # PROV relationships methods

    def _wasAttributedTo(self, entity, agent):
        self.graph.add((entity, PROV.wasAttributedTo, agent))

    def _wasAssociatedWith(self, activity, agent):
        self.graph.add((activity, PROV.wasAssociatedWith, agent))

    def _wasDerivedFrom(self, used_entity, generated_entity):
        self.graph.add((generated_entity, PROV.wasDerivedFrom, used_entity))

    def _wasGeneratedBy(self, entity, activity):
        self.graph.add((entity, PROV.wasGeneratedBy, activity))

    def _used(self, activity, entity):
        self.graph.add((activity, PROV.used, entity))

    # Agent methods

    def _add_ScriptAgent(self, script_info, session_id):
        # Adds a ScriptAgent record from the Alpaca PROV model
        uri = URIRef(script_identifier(script_info, session_id,
                                       self._authority))
        self.graph.add((uri, RDF.type, ALPACA.ScriptAgent))
        self.graph.add((uri, ALPACA.scriptPath, Literal(script_info.path)))
        return uri

    # Activity methods

    def _add_Function(self, function_info):
        # Adds a Function record from the Alpaca PROV model
        uri = URIRef(function_identifier(function_info, self._authority))
        self.graph.add((uri, RDF.type, ALPACA.Function))
        self.graph.add((uri, ALPACA.functionName,
                        Literal(function_info.name)))
        self.graph.add((uri, ALPACA.implementedIn,
                        Literal(function_info.module)))
        self.graph.add((uri, ALPACA.functionVersion,
                        Literal(function_info.version)))
        return uri

    def _add_FunctionExecution(self, script_info, session_id, execution_id,
                               function_info, params, execution_order,
                               code_statement, start, end, function):
        # Adds a FunctionExecution record from the Alpaca PROV model
        uri = URIRef(execution_identifier(
            script_info, function_info, session_id, execution_id,
            self._authority))
        self.graph.add((uri, RDF.type, ALPACA.FunctionExecution))
        self.graph.add((uri, PROV.startedAtTime,
                        Literal(start, datatype=XSD.dateTime)))
        self.graph.add((uri, PROV.endedAtTime,
                        Literal(end, datatype=XSD.dateTime)))
        self.graph.add((uri, ALPACA.codeStatement, Literal(code_statement)))
        self.graph.add((uri, ALPACA.executionOrder,
                        Literal(execution_order, datatype=XSD.integer)))
        self.graph.add((uri, ALPACA.usedFunction, function))

        for name, value in params.items():
            value = _ensure_type(value)
            _add_name_value_pair(self.graph, uri, ALPACA.hasParameter,
                                 name, value)
        return uri

    # Entity methods

    def _add_DataObjectEntity(self, info):
        # Adds a DataObjectEntity from the Alpaca PROV model
        # If the entity already exists, skip it
        uri = URIRef(data_object_identifier(info, self._authority))

        if uri in self.graph.subjects(RDF.type, ALPACA.DataObjectEntity):
            return uri
        self.graph.add((uri, RDF.type, ALPACA.DataObjectEntity))
        self.graph.add((uri, ALPACA.hashSource, Literal(info.hash_method)))
        self._add_entity_metadata(uri, info)
        return uri

    def _add_FileEntity(self, info):
        # Adds a FileEntity from the Alpaca PROV model
        uri = URIRef(file_identifier(info, self._authority))
        self.graph.add((uri, RDF.type, ALPACA.FileEntity))
        self.graph.add((uri, ALPACA.filePath,
                        Literal(info.path, datatype=XSD.string)))
        return uri

    def _add_entity_metadata(self, uri, info):
        # Add data object metadata (attributes, annotations) to the entities,
        # using properties from the Alpaca PROV model
        package_name = info.type.split(".")[0]
        metadata = info.details

        if package_name in self._metadata_plugins:
            # Handle objects like Neo objects (i.e., to avoid dumping all the
            # information in collections such as `segments` or `events`)
            self._metadata_plugins[package_name](self.graph, uri, metadata)
        else:
            # Add metadata using default handling, i.e., all attributes
            for name, value in metadata.items():
                # Make sure that types such as list and Quantity are handled
                value = _ensure_type(value)

                _add_name_value_pair(self.graph, uri=uri,
                                     predicate=ALPACA.hasAttribute,
                                     name=name,
                                     value=value)

    def _add_membership(self, container, child, params):
        # Add membership relationships according to the standard PROV model
        # and properties specific to the Alpaca PROV model
        predicates = {
            'name': ALPACA.fromAttribute,
            'index': ALPACA.containerIndex,
            'slice': ALPACA.containerSlice,
        }

        for name, value in params.items():
            predicate = predicates[name]
            self.graph.add((child, predicate, Literal(value)))
        self.graph.add((container, PROV.hadMember, child))

    def _create_entity(self, info):
        # Create an Alpaca PROV Entity based on DataObject/File information
        if isinstance(info, DataObject):
            return self._add_DataObjectEntity(info)
        elif isinstance(info, File):
            return self._add_FileEntity(info)
        raise ValueError("Invalid entity!")

    # Interface methods

    def _add_function_execution(self, execution, script_agent, script_info,
                                session_id):
        # Add one `FunctionExecution` record to the file, and generate all the
        # provenance semantic relationships

        def _is_membership(function_info):
            name = function_info.name
            return name in ("attribute", "subscript")

        function_info = execution.function
        if _is_membership(function_info):
            # attributes and subscripting operations
            container = execution.input[0]
            child = execution.output[0]
            container_entity = self._create_entity(container)
            if PROV.wasAttributedTo not in \
                    self.graph.predicates(container_entity, script_agent):
                self._wasAttributedTo(container_entity, script_agent)
            child_entity = self._create_entity(child)
            self._add_membership(container_entity, child_entity,
                                 execution.params)
        else:
            # This is a function execution. Add Function activity
            cur_function = self._add_Function(function_info)

            # Get the FunctionExecution node with function parameters and
            # other provenance info
            cur_activity = self._add_FunctionExecution(
                script_info=script_info, session_id=session_id,
                execution_id=execution.execution_id,
                function_info=function_info, params=execution.params,
                execution_order=execution.order,
                code_statement=execution.code_statement,
                start=execution.time_stamp_start,
                end=execution.time_stamp_end,
                function=cur_function,
            )

            # Add all the inputs as entities, and create a `used` association
            # with the activity. URNs differ when the input is a file or
            # Python object.
            input_entities = []
            for key, value in execution.input.items():
                cur_entities = []

                if isinstance(value, Container):
                    # If this is a Container, several objects are inside.
                    for element in value.elements:
                        cur_entities.append(self._create_entity(element))
                else:
                    cur_entities.append(self._create_entity(value))

                input_entities.extend(cur_entities)

                for cur_entity in cur_entities:
                    self._used(activity=cur_activity, entity=cur_entity)
                    self._wasAttributedTo(entity=cur_entity,
                                          agent=script_agent)

            # Add all the outputs as entities, and create the `wasGenerated`
            # relationship.
            output_entities = []
            for key, value in execution.output.items():
                cur_entity = self._create_entity(value)
                output_entities.append(cur_entity)
                self._wasGeneratedBy(entity=cur_entity, activity=cur_activity)
                self._wasAttributedTo(entity=cur_entity, agent=script_agent)

            # Iterate over the input/output pairs to add the `wasDerived`
            # relationship
            for input_entity, output_entity in \
                    product(input_entities, output_entities):
                self._wasDerivedFrom(used_entity=input_entity,
                                     generated_entity=output_entity)

            # Associate the activity to the script
            self._wasAssociatedWith(activity=cur_activity, agent=script_agent)

    def add_history(self, script_info, session_id, history):
        """
        Adds a history of `FunctionExecution` records captured by Alpaca to a
        PROV document using the Alpaca PROV ontology. The script is added as
        a `ScriptAgent` agent.

        Parameters
        ----------
        script_info : alpaca_types.File
            Named tuple with the information on the script being tracked
            (hash and file path).
        session_id : str
            Unique identifier for this script execution.
        history : list of FunctionExecution
            Provenance history to be serialized as PROV.
        """
        script_agent = self._add_ScriptAgent(script_info, session_id)
        for execution in history:
            self._add_function_execution(execution, script_agent, script_info,
                                         session_id)

    def read_records(self, file_name, file_format='turtle'):
        """
        Reads PROV data that was previously serialized as RDF.

        Parameters
        ----------
        file_name : str or Path-like
            Location of the file with PROV data to be read.
        file_format : {'json-ld', 'n3', 'nt', 'turtle', 'xml', 'ttl', 'rdf', 'json'}
            Format used to serialize the file that is being read. If None, the
            format will be inferred from the extension. The formats are
            the ones accepted by RDFLib. Some shortucts are defined for common
            file extensions:

            * 'ttl': Turtle
            * 'rdf': RDF-XML
            * 'json': JSON-LD

        Raises
        ------
        ValueError
            If `file_format` is None and `file_name` has no extension to infer
            the format, if it could not be inferred, or if the format is
            invalid.
        """
        if file_format is None:
            file_format = _get_prov_file_format(file_name)

        if file_format is None:
            raise ValueError("Could not infer serialization format. Please,"
                             "provide it explicitly using `file_format`.")

        if file_format not in ['json-ld', 'n3', 'nt', 'turtle', 'xml']:
            raise ValueError("Unsupported serialization format")

        with open(file_name, "r") as source:
            self.graph.parse(source, format=file_format)

    def serialize(self, file_name, file_format='turtle'):
        """
        Writes PROV data to a file or gets an in-memory string.

        Parameters
        ----------
        file_name : str or Path-like
            Location of the file with PROV data to be read.
        file_format : {'json-ld', 'n3', 'nt', 'hext', 'pretty-xml', 'trig', 'turtle', 'longturtle', 'xml'}
            Format used in the file that is being read. The format strings are
            the ones supported by RDFLib.
            If None, the format will be inferred from the extension.
            Default: 'turtle'

        """
        return self.graph.serialize(file_name, format=file_format)
